#!/usr/bin/env python3


def ax_extend(ax, x_extend_ratio=1.0, y_extend_ratio=1.0):
    ## Original coordinates
    bbox = ax.get_position()
    left_orig = bbox.x0
    bottom_orig = bbox.y0
    width_orig = bbox.x1 - bbox.x0
    height_orig = bbox.y1 - bbox.y0
    g_orig = (left_orig + width_orig / 2.0, bottom_orig + height_orig / 2.0)

    ## Target coordinates
    g_tgt = g_orig
    width_tgt = width_orig * x_extend_ratio
    height_tgt = height_orig * y_extend_ratio
    left_tgt = g_tgt[0] - width_tgt / 2
    bottom_tgt = g_tgt[1] - height_tgt / 2

    ax.set_position(
        [
            left_tgt,
            bottom_tgt,
            width_tgt,
            height_tgt,
        ]
    )
    return ax


# """
# ====================
# Demo Fixed Size Axes
# ====================
# """

# import matplotlib.pyplot as plt
# from mpl_toolkits.axes_grid1 import Divider, Size

# ###############################################################################


# fig = plt.figure(figsize=(6, 6))

# # The first items are for padding and the second items are for the axes.
# # sizes are in inch.
# h = [Size.Fixed(1.0), Size.Fixed(4.5)]
# v = [Size.Fixed(0.7), Size.Fixed(5.0)]

# divider = Divider(fig, (0, 0, 1, 1), h, v, aspect=False)
# # The width and height of the rectangle are ignored.

# ax = fig.add_axes(divider.get_position(), axes_locator=divider.new_locator(nx=1, ny=1))

# ax.plot([1, 2, 3])
# fig.show()

# ###############################################################################


# fig = plt.figure(figsize=(6, 6))

# # The first & third items are for padding and the second items are for the
# # axes. Sizes are in inches.
# h = [Size.Fixed(1.0), Size.Scaled(1.5), Size.Fixed(0.8)]
# v = [Size.Fixed(0.7), Size.Scaled(1.5), Size.Fixed(0.3)]

# divider = Divider(fig, (0, 0, 1, 1), h, v, aspect=False)
# # The width and height of the rectangle are ignored.

# ax = fig.add_axes(divider.get_position(), axes_locator=divider.new_locator(nx=1, ny=1))

# ax.plot([1, 2, 3])

# plt.show()


# def ax_extend(
#     fig,
#     ax,
#     x_extend_ratio=1.00,
#     y_extend_ratio=1.00,
# ):

#     ################################################################################
#     ## delete here
#     import matplotlib.pyplot as plt
#     import numpy as np

#     fig, ax = plt.subplots()
#     x_extend_ratio = 1.00
#     y_extend_ratio = 1.00
#     ################################################################################

#     bbox = ax.get_position()

#     ## Calculates delta ratios
#     fig_width_inch, fig_height_inch = fig.get_size_inches()

#     # x_delta_offset_inch = float(x_delta_offset_cm) / 2.54
#     # y_delta_offset_inch = float(y_delta_offset_cm) / 2.54

#     # x_delta_offset_ratio = x_delta_offset_inch / fig_width_inch
#     # y_delta_offset_ratio = y_delta_offset_inch / fig_width_inch

#     ## Determines updated bbox position
#     left = bbox.x0 + x_delta_offset_ratio
#     bottom = bbox.y0 + y_delta_offset_ratio
#     width = bbox.x1 - bbox.x0
#     height = bbox.y1 - bbox.y0

#     if dragh:
#         width -= x_delta_offset_ratio

#     if dragv:
#         height -= y_delta_offset_ratio

#     ax.set_position(
#         [
#             left,
#             bottom,
#             width,
#             height,
#         ]
#     )

#     return ax


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    fig, axes = plt.subplots(1, 2)
    ax = axes[1]
    ax = ax_extend(ax, 0.75, 1.1)
    fig.show()



#!/usr/bin/env python


import matplotlib


def ax_set_n_ticks(
    ax,
    n_xticks=4,
    n_yticks=4,
):
    """
    Example:
        ax = ax_set_n_ticks(ax)
    """

    ax.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(n_xticks))
    ax.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(n_yticks))

    return ax



#!/usr/bin/env python3
# Time-stamp: "2021-09-20 23:09:06 (ywatanabe)"

from . import dsp
from . import general
from . import ml
from . import plt
from . import stats
from . import resource


__copyright__ = "Copyright (C) 2021 Yusuke Watanabe"
__version__ = "0.0.15"
__license__ = "GPL3.0"
__author__ = "ywatanabe1989"
__author_email__ = "ywata1989@gmail.com"
__url__ = "https://github.com/ywatanabe1989/mngs"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2021-09-20 23:07:01 (ywatanabe)"

import psutil
import platform
from datetime import datetime
import pandas as pd


get_size = mngs.general.fmt_size
# def size(bytes, suffix="B"):
#     """
#     Scale bytes to its proper format
#     e.g:
#         1253656 => '1.20MB'
#         1253656678 => '1.17GB'
#     """
#     factor = 1024
#     for unit in ["", "K", "M", "G", "T", "P"]:
#         if bytes < factor:
#             return f"{bytes:.2f}{unit}{suffix}"
#         bytes /= factor


def system():
    print("=" * 40, "System Information", "=" * 40)
    uname = platform.uname()
    print(f"System: {uname.system}")
    print(f"Node Name: {uname.node}")
    print(f"Release: {uname.release}")
    print(f"Version: {uname.version}")
    print(f"Machine: {uname.machine}")
    print(f"Processor: {uname.processor}")


def boot_time():
    # Boot Time
    print("=" * 40, "Boot Time", "=" * 40)
    boot_time_timestamp = psutil.boot_time()
    bt = datetime.fromtimestamp(boot_time_timestamp)
    print(f"Boot Time: {bt.year}/{bt.month}/{bt.day} {bt.hour}:{bt.minute}:{bt.second}")


def cpu():
    # let's print CPU information
    print("=" * 40, "CPU Info", "=" * 40)
    # number of cores
    print("Physical cores:", psutil.cpu_count(logical=False))
    print("Total cores:", psutil.cpu_count(logical=True))
    # CPU frequencies
    cpufreq = psutil.cpu_freq()
    print(f"Max Frequency: {cpufreq.max:.2f}Mhz")
    print(f"Min Frequency: {cpufreq.min:.2f}Mhz")
    print(f"Current Frequency: {cpufreq.current:.2f}Mhz")
    # CPU usage
    print("CPU Usage Per Core:")
    for i, percentage in enumerate(psutil.cpu_percent(percpu=True, interval=1)):
        print(f"Core {i}: {percentage}%")
    print(f"Total CPU Usage: {psutil.cpu_percent()}%")


def memory():
    # Memory Information
    print("=" * 40, "Memory Information", "=" * 40)
    # get the memory details
    svmem = psutil.virtual_memory()
    print(f"Total: {get_size(svmem.total)}")
    print(f"Available: {get_size(svmem.available)}")
    print(f"Used: {get_size(svmem.used)}")
    print(f"Percentage: {svmem.percent}%")
    print("=" * 20, "SWAP", "=" * 20)
    # get the swap memory details (if exists)
    swap = psutil.swap_memory()
    print(f"Total: {get_size(swap.total)}")
    print(f"Free: {get_size(swap.free)}")
    print(f"Used: {get_size(swap.used)}")
    print(f"Percentage: {swap.percent}%")


def RAM():
    return memory()


def disk():
    # Disk Information
    print("=" * 40, "Disk Information", "=" * 40)
    print("Partitions and Usage:")
    # get all disk partitions
    partitions = psutil.disk_partitions()
    for partition in partitions:
        print(f"=== Device: {partition.device} ===")
        print(f"  Mountpoint: {partition.mountpoint}")
        print(f"  File system type: {partition.fstype}")
        try:
            partition_usage = psutil.disk_usage(partition.mountpoint)
        except PermissionError:
            # this can be catched due to the disk that
            # isn't ready
            continue
        print(f"  Total Size: {get_size(partition_usage.total)}")
        print(f"  Used: {get_size(partition_usage.used)}")
        print(f"  Free: {get_size(partition_usage.free)}")
        print(f"  Percentage: {partition_usage.percent}%")
    # get IO statistics since boot
    disk_io = psutil.disk_io_counters()
    print(f"Total read: {get_size(disk_io.read_bytes)}")
    print(f"Total write: {get_size(disk_io.write_bytes)}")


def network():
    # Network information
    print("=" * 40, "Network Information", "=" * 40)
    # get all network interfaces (virtual and physical)
    if_addrs = psutil.net_if_addrs()
    for interface_name, interface_addresses in if_addrs.items():
        for address in interface_addresses:
            print(f"=== Interface: {interface_name} ===")
            if str(address.family) == "AddressFamily.AF_INET":
                print(f"  IP Address: {address.address}")
                print(f"  Netmask: {address.netmask}")
                print(f"  Broadcast IP: {address.broadcast}")
            elif str(address.family) == "AddressFamily.AF_PACKET":
                print(f"  MAC Address: {address.address}")
                print(f"  Netmask: {address.netmask}")
                print(f"  Broadcast MAC: {address.broadcast}")
    # get IO statistics since boot
    net_io = psutil.net_io_counters()
    print(f"Total Bytes Sent: {get_size(net_io.bytes_sent)}")
    print(f"Total Bytes Received: {get_size(net_io.bytes_recv)}")


if __name__ == "__main__":
    system()
    boot_time()
    cpu()
    memory()
    disk()
    network()



#!/usr/bin/env python3

import os
import sys

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import mngs
import yaml
from natsort import natsorted
from sklearn.metrics import (
    balanced_accuracy_score,
    classification_report,
    confusion_matrix,
    matthews_corrcoef,
    roc_auc_score,
)


class Reporter:
    """Saves the following metrics under sdir.
       - Balanced Accuracy
       - MCC
       - Confusion Matrix
       - Classification Report
       - ROC AUC score / curve
       - PRE-REC AUC score / curve

    Manual adding example:
        ##############################
        ## fig
        ##############################
        fig, ax = plt.subplots()
        ax.plot(np.random.rand(10))
        reporter.add(
            "manu_figs",
            fig,
            {
                "dirname": "manu_fig_dir/",
                "ext": ".png",
            },
        )
        ##############################
        ## DataFrame
        ##############################
        df = pd.DataFrame(np.random.rand(5, 3))
        reporter.add("manu_dfs", df, {"fname": "manu_dfs.csv", "method": "mean"})
        ##############################
        ## scalar
        ##############################
        scalar = random.random()
        reporter.add(
            "manu_scalars",
            scalar,
            {"fname": "manu_scalars.csv", "column_name": "manu_column_name"},
        )
    """

    def __init__(self, sdir):
        self.sdir = sdir
        self.ts = mngs.general.TimeStamper()

        self.balanced_accs_folds = []
        self.mccs_folds = []
        self.conf_mats_folds = []
        self.clf_reports_folds = []
        self.roc_aucs_macro_folds = []
        self.roc_aucs_micro_folds = []
        self.roc_aucs_figs_folds = []
        self.pre_rec_aucs_macro_folds = []
        self.pre_rec_aucs_micro_folds = []
        self.pre_rec_aucs_figs_folds = []

        ## for more flexible collection
        self.added_folds_obj_dict = mngs.general.listed_dict()
        self.added_folds_meta_dict = {}
        self._added_at_least_once = False

        print("\n{}\n".format(mngs.general.gen_timestamp()))
        self.ts("\nReporter has been initialized.\n")

    def add(
        self,
        obj_name,
        obj,
        meta_dict,
    ):

        ## Initializes the list
        if not self._added_at_least_once:
            self.added_folds_obj_dict[obj_name] = []
        self._added_at_least_once = True

        ## adds
        self.added_folds_obj_dict[obj_name].append(obj)
        self.added_folds_meta_dict[obj_name] = meta_dict

    def calc_metrics(
        self,
        true_class,
        pred_class,
        pred_proba,
        labels=None,
        i_fold=None,
        show=True,
        plot=False,
    ):
        """Calculates ACC, Confusion Matrix, Classification Report, and ROC-AUC score."""
        self.labels = labels

        true_class, pred_class, pred_proba = (
            mngs.general.torch_to_arr(true_class),
            mngs.general.torch_to_arr(pred_class),
            mngs.general.torch_to_arr(pred_proba),
        )

        ####################
        ## ACC
        ####################
        acc = (true_class.reshape(-1) == pred_class.reshape(-1)).mean()
        if show:
            print("\nACC in fold#{} was {:.3f}\n".format(i_fold, acc))
        balanced_acc = balanced_accuracy_score(
            true_class.reshape(-1), pred_class.reshape(-1)
        )
        if show:
            print("\nBalanced ACC in fold#{} was {:.3f}\n".format(i_fold, balanced_acc))

        ####################
        ## MCC
        ####################
        mcc = matthews_corrcoef(true_class.reshape(-1), pred_class.reshape(-1))
        if show:
            print("\nMCC in fold#{} was {:.3f}\n".format(i_fold, mcc))

        ####################
        ## Confusion Matrix
        ####################
        conf_mat = confusion_matrix(true_class, pred_class)
        conf_mat = pd.DataFrame(data=conf_mat, columns=labels).set_index(
            pd.Series(list(labels))
        )
        if show:
            print("\nConfusion Matrix in fold#{}: \n{}\n".format(i_fold, conf_mat))

        ####################
        ## Classification Report
        ####################
        clf_report = pd.DataFrame(
            classification_report(
                true_class,
                pred_class,
                target_names=labels,
                output_dict=True,
            )
        )
        # accuracy to balanced accuracy
        clf_report["accuracy"] = balanced_acc
        clf_report = clf_report.rename(columns={"accuracy": "balanced accuracy"})
        clf_report = clf_report.round(3)
        # rename 'support' to 'sample size'
        clf_report["index"] = clf_report.index
        clf_report.loc["support", "index"] = "sample size"
        clf_report.set_index("index", drop=True, inplace=True)
        clf_report.index.name = None
        if show:
            print(
                "\nClassification Report in fold#{}: \n{}\n".format(i_fold, clf_report)
            )

        ####################
        ## ROC-AUC score
        ####################
        mngs.plt.configure_mpl(
            plt,
            figsize=(7, 7),
            labelsize=8,
            fontsize=7,
            legendfontsize=6,
            tick_size=0.8,
            tick_width=0.2,
        )

        fig_roc, metrics_roc_auc_dict = mngs.ml.plt.roc_auc(
            plt,
            true_class,
            pred_proba,
            labels,
        )
        if plot:
            fig_roc.plot()
        else:
            plt.close()

        ####################
        ## PRE-REC AUC score
        ####################
        fig_pre_rec, metrics_pre_rec_auc_dict = mngs.ml.plt.pre_rec_auc(
            plt, true_class, pred_proba, labels
        )
        if plot:
            fig_pre_rec.plot()
        else:
            plt.close()

        ####################
        # to buffer
        ####################
        self.mccs_folds.append(mcc)
        self.balanced_accs_folds.append(balanced_acc)
        self.conf_mats_folds.append(conf_mat)
        self.clf_reports_folds.append(clf_report)
        self.roc_aucs_micro_folds.append(metrics_roc_auc_dict["roc_auc"]["micro"])
        self.roc_aucs_macro_folds.append(metrics_roc_auc_dict["roc_auc"]["macro"])
        self.roc_aucs_figs_folds.append(fig_roc)
        self.pre_rec_aucs_micro_folds.append(
            metrics_pre_rec_auc_dict["pre_rec_auc"]["micro"]
        )
        self.pre_rec_aucs_macro_folds.append(
            metrics_pre_rec_auc_dict["pre_rec_auc"]["macro"]
        )
        self.pre_rec_aucs_figs_folds.append(fig_pre_rec)
        del fig_roc, fig_pre_rec

    def summarize(
        self,
    ):
        ####################
        ## Summarize each fold's metirics
        ####################
        self.mcc_cv_mean, self.mcc_cv_std = self.take_mean_and_std(self.mccs_folds)
        self.balanced_acc_cv_mean, self.balanced_acc_cv_std = self.take_mean_and_std(
            self.balanced_accs_folds
        )
        self.conf_mat_cv_sum = self.summarize_dfs(self.conf_mats_folds, method="sum")
        self.clf_report_cv_mean, self.clf_report_cv_std = self.summarize_dfs(
            self.clf_reports_folds, method="mean"
        )
        self.roc_auc_micro_cv_mean, self.roc_auc_micro_cv_std = self.take_mean_and_std(
            self.roc_aucs_micro_folds
        )
        self.roc_auc_macro_cv_mean, self.roc_auc_macro_cv_std = self.take_mean_and_std(
            self.roc_aucs_macro_folds
        )
        (
            self.pre_rec_auc_micro_cv_mean,
            self.pre_rec_auc_micro_cv_std,
        ) = self.take_mean_and_std(self.pre_rec_aucs_micro_folds)
        (
            self.pre_rec_auc_macro_cv_mean,
            self.pre_rec_auc_macro_cv_std,
        ) = self.take_mean_and_std(self.pre_rec_aucs_macro_folds)

        self.num_folds = len(self.conf_mats_folds)
        self.print_metrics()

    def print_metrics(self):
        print("\n --- {}-fold CV overall metrics --- \n".format(self.num_folds))
        print(
            "\nThe Mattews correlation coefficient: {} +/- {} (mean +/- std.; n={})\n".format(
                self.mcc_cv_mean, self.mcc_cv_std, self.num_folds
            )
        )

        print(
            "\nBalanced Accuracy Score: {} +/- {} (mean +/- std.; n={})\n".format(
                self.balanced_acc_cv_mean, self.balanced_acc_cv_std, self.num_folds
            )
        )

        print(
            "\nConfusion Matrix (Test; sum; num. folds={})\n{}\n".format(
                self.num_folds, self.conf_mat_cv_sum
            )
        )
        print(
            "\nClassification Report (Test; mean; num. folds={})\n{}\n".format(
                self.num_folds, self.clf_report_cv_mean
            )
        )
        print(
            "\nClassification Report (Test; std; num. folds={})\n{}\n".format(
                self.num_folds, self.clf_report_cv_std
            )
        )
        print(
            "\nROC AUC micro Score: {} +/- {} (mean +/- std.; n={})\n".format(
                self.roc_auc_micro_cv_mean, self.roc_auc_micro_cv_std, self.num_folds
            )
        )
        print(
            "\nROC AUC macro Score: {} +/- {} (mean +/- std.; n={})\n".format(
                self.roc_auc_macro_cv_mean, self.roc_auc_macro_cv_std, self.num_folds
            )
        )
        print(
            "\nPrecision-Recall AUC micro Score: {} +/- {} (mean +/- std.; n={})\n".format(
                self.pre_rec_auc_micro_cv_mean,
                self.pre_rec_auc_micro_cv_std,
                self.num_folds,
            )
        )
        print(
            "\nPrecision-Recall AUC macro Score: {} +/- {} (mean +/- std.; n={})\n".format(
                self.pre_rec_auc_macro_cv_mean,
                self.pre_rec_auc_macro_cv_std,
                self.num_folds,
            )
        )

    def take_mean_and_std(self, obj_list, n_round=3):
        arr = np.array(obj_list)
        return arr.mean(axis=0).round(n_round), arr.std(axis=0).round(n_round)

    def save_listed_scalars(
        self,
        listed_scalars,
        fname=None,
        column_name=None,
        n_round=3,
        show=True,
        makedirs=False,
    ):
        cat = [
            np.mean(listed_scalars).round(n_round),
            np.std(listed_scalars).round(n_round),
        ] + listed_scalars

        num_folds = len(listed_scalars)

        indi_suffix_cat = (
            ["{}-fold CV mean".format(num_folds)]
            + ["{}-fold CV std.".format(num_folds)]
            + ["fold#{}".format(i) for i in range(num_folds)]
        )

        mngs.general.save(
            cat,
            self.sdir + fname,
            indi_suffix=indi_suffix_cat,
            column_name=column_name,
            show=show,
            makedirs=makedirs,
        )

    def save_listed_dfs(
        self,
        listed_dfs,
        method="mean",
        fname=None,
        n_round=3,
        show=True,
        makedirs=False,
    ):
        if method == "mean":
            summarized_dfs = self.summarize_dfs(
                listed_dfs, method=method, n_round=n_round
            )
            indi_suffix_cat = (
                ["{}-fold CV mean".format(self.num_folds)]
                + ["{}-fold CV std.".format(self.num_folds)]
                + ["fold#{}".format(i) for i in range(self.num_folds)]
            )

        if method == "sum":
            summarized_dfs = self.summarize_dfs(
                listed_dfs, method=method, n_round=n_round
            )
            summarized_dfs = [summarized_dfs]
            indi_suffix_cat = ["sum-fold CV mean".format(self.num_folds)] + [
                "fold#{}".format(i) for i in range(self.num_folds)
            ]

        cat = [*summarized_dfs] + listed_dfs

        mngs.general.save(
            cat,
            self.sdir + fname,
            indi_suffix=indi_suffix_cat,
            overwrite=True,
            show=show,
            makedirs=makedirs,
        )

    def summarize_dfs(self, dfs_list, method="sum", n_round=3):
        df_zero = 0 * dfs_list[0].copy()  # get the table format

        if method == "sum":
            return df_zero + np.array(dfs_list).sum(axis=0).round(n_round)

        if method == "mean":
            arr = np.array(dfs_list)
            df_mean = arr.astype(float).mean(axis=0).round(n_round)
            df_std = arr.astype(float).std(axis=0).round(n_round)
            return df_zero + df_mean, df_zero + df_std

    def save_listed_figures(
        self, listed_figures, dirname="None", ext="png", show=True, makedirs=False
    ):
        for i_fold, obj in enumerate(listed_figures):  # range(len(listed_figures)):
            spath = self.sdir + dirname + "fold#{}".format(i_fold) + ext
            mngs.general.save(obj, spath, show=show, makedirs=makedirs)
        # print("\nSaved to: {s}\n".format(s=dirname))

    def save(self, others_dict=None, makedirs=True):
        # if makedirs:
        #     os.makedirs(self.sdir, exist_ok=True)

        ####################
        ## Others dict
        ####################
        if isinstance(others_dict, dict):
            for sfname, obj in others_dict.items():
                mngs.general.save(obj, self.sdir + sfname, makedirs=makedirs)

        ####################
        ## MCC
        ####################
        self.save_listed_scalars(
            self.mccs_folds,
            fname="mccs.csv",
            column_name="The Mattews corr. coeff.",
            n_round=3,
            makedirs=makedirs,
        )

        ####################
        ## Balanced ACC
        ####################
        self.save_listed_scalars(
            self.balanced_accs_folds,
            fname="balanced_accs.csv",
            column_name="Balanced Accuracy",
            n_round=3,
            makedirs=makedirs,
        )

        ####################
        ## Confusion Matrix
        ####################
        mngs.plt.configure_mpl(
            plt,
            figsize=(8, 8),
            labelsize=8,
            fontsize=6,
            legendfontsize=6,
            tick_size=0.8,
            tick_width=0.2,
        )

        # each fold
        self.conf_mats_folds_fig = [
            mngs.ml.plt.confusion_matrix(
                plt,
                cm,
                labels=self.labels,
                title="Test Mouse#{}".format(i_cm + 1),
                extend_ratio=0.4,
                colorbar=True,
            )
            for i_cm, cm in enumerate(self.conf_mats_folds)
        ]

        for fig_cm_fold in self.conf_mats_folds_fig:
            fig_cm_fold.axes[-1] = mngs.plt.ax_scientific_notation(
                fig_cm_fold.axes[-1],
                3,
                fformat="%3.1f",
                y=True,
            )

        for i_fold, cm in enumerate(self.conf_mats_folds_fig):
            mngs.general.save(
                cm, self.sdir + "conf_mat/" + "fold#{}.png".format(i_fold)
            )
        self.save_listed_dfs(
            self.conf_mats_folds,
            method="sum",
            fname="conf_mat/conf_mats.csv",
            makedirs=makedirs,
        )
        # kCV overall-sum
        title = "{}-CV overall-sum".format(len(self.conf_mats_folds))

        fig_cm = mngs.ml.plt.confusion_matrix(
            plt,
            self.conf_mat_cv_sum,
            labels=self.labels,
            title=title,
            extend_ratio=0.4,
            colorbar=True,
        )

        # fixme
        fig_cm.axes[-1] = mngs.plt.ax_scientific_notation(
            fig_cm.axes[-1],
            3,
            fformat="%3.1f",
            y=True,
        )

        mngs.general.save(
            fig_cm, self.sdir + "conf_mat/overall_sum.png", makedirs=makedirs
        )

        ####################
        ## Classification Report
        ####################
        self.save_listed_dfs(
            self.clf_reports_folds,
            method="mean",
            fname="clf_reports.csv",
            makedirs=makedirs,
        )

        ####################
        ## ROC-AUC and PRE-REC AUC
        ####################
        _roc_aucs_micro_cat = [
            self.roc_auc_micro_cv_mean,
            self.roc_auc_micro_cv_std,
        ] + self.roc_aucs_micro_folds
        _roc_aucs_macro_cat = [
            self.roc_auc_macro_cv_mean,
            self.roc_auc_macro_cv_std,
        ] + self.roc_aucs_macro_folds
        _pre_rec_aucs_micro_cat = [
            self.pre_rec_auc_micro_cv_mean,
            self.pre_rec_auc_micro_cv_std,
        ] + self.pre_rec_aucs_micro_folds
        _pre_rec_aucs_macro_cat = [
            self.pre_rec_auc_macro_cv_mean,
            self.pre_rec_auc_macro_cv_std,
        ] + self.pre_rec_aucs_macro_folds
        _indi_suffix_cat = (
            ["{}-fold CV mean".format(self.num_folds)]
            + ["{}-fold CV std.".format(self.num_folds)]
            + ["fold#{}".format(i) for i in range(self.num_folds)]
        )

        aucs_df = pd.DataFrame(
            {
                "ROC AUC micro": _roc_aucs_micro_cat,
                "ROC AUC macro": _roc_aucs_macro_cat,
                "PRE-REC AUC micro": _pre_rec_aucs_micro_cat,
                "PRE-REC AUC macro": _pre_rec_aucs_macro_cat,
            }
        )
        aucs_df.index = _indi_suffix_cat
        mngs.general.save(aucs_df, self.sdir + "aucs.csv", makedirs=makedirs)

        for i_fold in range(len(self.roc_aucs_figs_folds)):
            spath = self.sdir + "roc_curves/fold#{}.png".format(i_fold)
            mngs.general.save(
                self.roc_aucs_figs_folds[i_fold], spath, makedirs=makedirs
            )

        for i_fold in range(len(self.pre_rec_aucs_figs_folds)):
            spath = self.sdir + "pre_rec_curves/fold#{}.png".format(i_fold)
            mngs.general.save(
                self.pre_rec_aucs_figs_folds[i_fold], spath, makedirs=makedirs
            )

        ## Manually added scalars, dfs, and figures
        for k in self.added_folds_obj_dict.keys():
            listed_obj = self.added_folds_obj_dict[k]
            meta = self.added_folds_meta_dict[k]

            if mngs.general.is_listed_X(listed_obj, [int, float]):
                self.save_listed_scalars(
                    listed_obj, show=True, makedirs=makedirs, **meta
                )

            if mngs.general.is_listed_X(listed_obj, pd.DataFrame):
                self.save_listed_dfs(listed_obj, show=True, makedirs=makedirs, **meta)

            # if is_listed_X(
            if mngs.general.is_listed_X(
                listed_obj,
                [plt.Figure, matplotlib.figure.Figure],
            ):
                self.save_listed_figures(
                    listed_obj, show=True, makedirs=makedirs, **meta
                )


if __name__ == "__main__":
    import random

    import numpy as np
    import mngs
    from catboost import CatBoostClassifier, Pool
    from sklearn.datasets import load_digits
    from sklearn.model_selection import StratifiedKFold

    mngs.general.fix_seeds(np=np)

    ## Loads
    mnist = load_digits()
    X, T = mnist.data, mnist.target
    labels = mnist.target_names.astype(str)

    ## Make save dir
    sdir = "/tmp/sdir/"

    skf = StratifiedKFold(n_splits=5, shuffle=True)

    ## Main
    reporter = Reporter(sdir)
    for i_fold, (indi_tra, indi_tes) in enumerate(skf.split(X, T)):
        print(i_fold)
        ## koko
        X_tra, T_tra = X[indi_tra], T[indi_tra]
        X_tes, T_tes = X[indi_tes], T[indi_tes]

        clf = CatBoostClassifier(verbose=False)

        clf.fit(X_tra, T_tra, verbose=False)

        ## Prediction
        pred_proba_tes = clf.predict_proba(X_tes)
        pred_cls_tes = np.argmax(pred_proba_tes, axis=1)

        ##############################
        # fig
        ##############################
        fig, ax = plt.subplots()
        ax.plot(np.arange(10))
        reporter.add(
            "manu_figs",
            fig,
            {
                "dirname": "manu_fig_dir/",
                "ext": ".png",
            },
        )

        df = pd.DataFrame(np.random.rand(5, 3))
        reporter.add("manu_dfs", df, {"fname": "manu_dfs.csv", "method": "mean"})

        # scalar
        scalar = random.random()
        reporter.add(
            "manu_scalars",
            scalar,
            {"fname": "manu_scalars.csv", "column_name": "manu_column_name"},
        )

        ## Metrics
        reporter.calc_metrics(
            T_tes,
            pred_cls_tes,
            pred_proba_tes,
            labels=labels,
            i_fold=i_fold,
        )

    reporter.summarize()
    reporter.save()

    ## EOF



#!/usr/bin/env python3

import math
import re
import time
from bisect import bisect_left, bisect_right
from collections import defaultdict

import numpy as np
import torch


################################################################################
## strings
################################################################################
def connect_strs(strs, filler="_"):  # connect_nums also works as connect_strs
    """
    Example:
        print(connect_strs(['a', 'b', 'c'], filler='_'))
        # 'a_b_c'
    """
    if isinstance(strs, list) or isinstance(strs, tuple):
        connected = ""
        for s in strs:
            connected += filler + s
        return connected[len(filler) :]
    else:
        return strs


def connect_nums(nums, filler="_"):
    """
    Example:
        print(connect_nums([1, 2, 3], filler='_'))
        # '1_2_3'
    """
    if isinstance(nums, list) or isinstance(nums, tuple):
        connected = ""
        for n in nums:
            connected += filler + str(n)
        return connected[len(filler) :]
    else:
        return nums


def squeeze_spaces(string, pattern=" +", repl=" "):
    """Return the string obtained by replacing the leftmost
    non-overlapping occurrences of the pattern in string by the
    replacement repl.  repl can be either a string or a callable;
    if a string, backslash escapes in it are processed.  If it is
    a callable, it's passed the Match object and must return
    a replacement string to be used.
    """
    # return re.sub(" +", " ", string)
    return re.sub(pattern, repl, string)


def grep(str_list, search_key):
    """
    Example:
        str_list = ['apple', 'orange', 'apple', 'apple_juice', 'banana', 'orange_juice']
        search_key = 'orange'
        print(grep(str_list, search_key))
        # ([1, 5], ['orange', 'orange_juice'])
    """
    matched_keys = []
    indi = []
    for ii, string in enumerate(str_list):
        m = re.search(search_key, string)
        if m is not None:
            matched_keys.append(string)
            indi.append(ii)
    return indi, matched_keys


def pop_keys(keys_list, keys_to_pop):
    """
    keys_list = ['a', 'b', 'c', 'd', 'e', 'bde']
    keys_to_pop = ['b', 'd']
    pop_keys(keys_list, keys_to_pop)
    """
    indi_to_remain = [k not in keys_to_pop for k in keys_list]
    keys_remainded_list = list(np.array(keys_list)[list(indi_to_remain)])
    return keys_remainded_list


def fmt_size(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)


################################################################################
## list
################################################################################
def take_the_closest(list_obj, num_insert):
    """
    Assumes list_obj is sorted. Returns the closest value to num.
    If the same number is included in list_obj, the smaller number is returned.

    Example:
        list_obj = np.array([0, 1, 1, 2, 3, 3])
        num = 1.2
        closest_num, closest_pos = take_the_closest(list_obj, num)
        print(closest_num, closest_pos)
        # 1 2

        list_obj = np.array([0, 1, 1, 2, 3, 3])
        num = 1
        closest_num, closest_pos = take_the_closest(list_obj, num)
        print(closest_num, closest_pos)
        # 1 1
    """
    if math.isnan(num_insert):
        closest_num = np.nan
        closest_pos = np.nan

    pos_num_insert = bisect_left(list_obj, num_insert)

    if pos_num_insert == 0:  # When the insertion is at the first position
        closest_num = list_obj[0]
        closest_pos = pos_num_insert  # 0

    if pos_num_insert == len(list_obj):  # When the insertion is at the last position
        closest_num = list_obj[-1]
        closest_pos = pos_num_insert  # len(list_obj)

    else:  # When the insertion is anywhere between the first and the last positions
        pos_before = pos_num_insert - 1

        before_num = list_obj[pos_before]
        after_num = list_obj[pos_num_insert]

        delta_after = abs(after_num - num_insert)
        delta_before = abs(before_num - num_insert)

        if np.abs(delta_after) < np.abs(delta_before):
            closest_num = after_num
            closest_pos = pos_num_insert

        else:
            closest_num = before_num
            closest_pos = pos_before

    return closest_num, closest_pos


################################################################################
## dictionary
################################################################################
def listed_dict(keys=None):  # Is there a better name?
    """
    Example 1:
        import random
        random.seed(42)
        d = listed_dict()
        for _ in range(10):
            d['a'].append(random.randint(0, 10))
        print(d)
        # defaultdict(<class 'list'>, {'a': [10, 1, 0, 4, 3, 3, 2, 1, 10, 8]})

    Example 2:
        import random
        random.seed(42)
        keys = ['a', 'b', 'c']
        d = listed_dict(keys)
        for _ in range(10):
            d['a'].append(random.randint(0, 10))
            d['b'].append(random.randint(0, 10))
            d['c'].append(random.randint(0, 10))
        print(d)
        # defaultdict(<class 'list'>, {'a': [10, 4, 2, 8, 6, 1, 8, 8, 8, 7],
        #                              'b': [1, 3, 1, 1, 0, 3, 9, 3, 6, 9],
        #                              'c': [0, 3, 10, 9, 0, 3, 0, 10, 3, 4]})
    """
    dict_list = defaultdict(list)
    # initialize with keys if possible
    if keys is not None:
        for k in keys:
            dict_list[k] = []
    return dict_list


################################################################################
## variables
################################################################################
def does_exist(suspicious_var_str):
    """
    Example:
        print(does_exist('a'))
        # False

        a = 5
        print(does_exist('a'))
        # True
    """
    return suspicious_var_str in globals()



#!/usr/bin/env python3

from pprint import pprint

# import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
import mngs


class LearningCurveLogger(object):
    def __init__(
        self,
    ):
        self.steps_str = ("Training", "Test")
        self.logged_dict = mngs.general.listed_dict(self.steps_str)
        for k in self.steps_str:
            self.logged_dict[k] = {}

    def __call__(self, log_dict, step="Training"):
        for k in log_dict.keys():
            try:
                self.logged_dict[step][k].append(log_dict[k])
            except:
                self.logged_dict[step][k] = []
                self.logged_dict[step][k].append(log_dict[k])

    def to_dfs(self, steps_str=["Training", "Test"]):  # fixme
        self.dfs = {}
        self.dfs_pivot = {}
        for step in steps_str:
            logged_dict = self.logged_dict[step]
            length = len(logged_dict["i_epoch"])
            df = pd.DataFrame(
                {
                    "step": [step for _ in range(length)],
                    "i_mouse_test": logged_dict["i_mouse_test"],
                    "i_epoch": logged_dict["i_epoch"],
                    "i_global": logged_dict["i_global"],
                    "loss": logged_dict["loss"],
                    "Balanced ACC": logged_dict["balanced_ACC"],
                    "pred_proba": [logged_dict["pred_proba"][i] for i in range(length)],
                    "gt_label": [logged_dict["gt_label"][i] for i in range(length)],
                }
            )  # fixme

            self.dfs[step] = df

            # pivot version to plot
            self.dfs_pivot[step] = (
                self.dfs[step][
                    [
                        "i_global",
                        "Balanced ACC",
                        "loss",
                    ]
                ]
                .pivot_table(columns=["i_global"], aggfunc="mean")
                .T
            )

    def print_learning_curves_in_digits(self, step="Training"):
        print(step)

        try:
            df = pd.DataFrame(
                {
                    "i_epoch": self.logged_dict[step]["i_epoch"],
                    "Balanced ACC": self.logged_dict[step]["balanced_ACC"],
                }
            ).pivot_table(columns=["i_epoch"], aggfunc="mean")
            pprint(df)
        except:
            pass

        print()

    def plot_learning_curves(
        self,
        plt,
        i_mouse_test=None,
        max_epochs=None,
        window_size_sec=None,
        max_n_ticks=4,
    ):

        self.to_dfs()

        ## Plot
        fig, ax = plt.subplots(2, 1, sharex=True, sharey=False)
        title = "fold#{}; MAX EPOCHS = {}; WINDOW SIZE = {} [sec]".format(
            i_mouse_test, max_epochs, window_size_sec
        )
        fig.text(0.5, 0.95, title, ha="center")

        keys_to_plot = ["loss", "Balanced ACC"]

        COLOR_DICT = {
            "Training": "blue",
            "Test": "green",
        }

        for i_plt, plt_k in enumerate(keys_to_plot):
            for step_k in self.dfs_pivot.keys():
                if step_k == "Training":  # line
                    ax[i_plt].plot(
                        self.dfs_pivot[step_k].index,
                        self.dfs_pivot[step_k][plt_k],
                        label=step_k,
                        color=COLOR_DICT[step_k],
                        linewidth=1,
                        # linewidth=3,
                    )

                if step_k == "Test":  # scatter
                    ax[i_plt].scatter(
                        self.dfs_pivot[step_k].index,
                        self.dfs_pivot[step_k][plt_k],
                        label=step_k,
                        color=COLOR_DICT[step_k],
                        # s=150,
                        s=50,
                        alpha=0.9,
                    )

                if "Balanced ACC" in plt_k:
                    ax[i_plt].set_ylim(0, 1)

        ax[0].legend()
        ax[1].legend()

        ax[1].set_xlabel("Iteration#")

        ax[0].set_ylabel("Loss")
        ax[1].set_ylabel("Balanced Accuracy")

        for a in ax:
            a.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_ticks))
            a.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_ticks))

        return fig


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import torch
    from scipy.special import softmax

    lc_logger = LearningCurveLogger()
    i_global = 0
    batch_size = 64
    n_classes = 2
    i_mouse_test = 0
    max_epochs = 3

    for i_epoch in range(max_epochs):
        step = "Training"
        for i_batch in enumerate(range(1000)):

            log_dict = {
                "loss": float(np.random.rand(1)),
                "balanced_ACC": float(np.random.rand(1)),
                "pred_proba": softmax(np.random.rand(batch_size, n_classes), axis=-1),
                "gt_label": np.random.randint(n_classes, size=batch_size),
                "i_mouse_test": i_mouse_test,
                "i_epoch": i_epoch,
                "i_global": i_global,
            }

            lc_logger(log_dict, step=step)

            i_global += 1

    step = "Test"
    for i_batch in enumerate(range(1000)):

        log_dict = {
            "loss": float(np.random.rand(1)),
            "balanced_ACC": float(np.random.rand(1)),
            "pred_proba": softmax(np.random.rand(batch_size, n_classes), axis=-1),
            "gt_label": np.random.randint(n_classes, size=batch_size),
            "i_mouse_test": i_mouse_test,
            "i_epoch": i_epoch,
            "i_global": i_global,
        }

        lc_logger(log_dict, step=step)

    mngs.plt.configure_mpl(
        plt,
        figsize=(8.7, 10),
        labelsize=8,
        fontsize=7,
        legendfontsize=6,
        tick_size=0.8,
        tick_width=0.2,
    )

    fig = lc_logger.plot_learning_curves(
        plt,
        i_mouse_test=i_mouse_test,
        max_epochs=max_epochs,
        window_size_sec=1024,
    )
    fig.show()



#!/usr/bin/env python3

import mngs


################################################################################
## Reproducibility
################################################################################
def fix_seeds(os=None, random=None, np=None, torch=None, tf=None, seed=42, show=True):
    # https://github.com/lucidrains/vit-pytorch/blob/main/examples/cats_and_dogs.ipynb
    if os is not None:
        import os

        os.environ["PYTHONHASHSEED"] = str(seed)

    if random is not None:
        random.seed(seed)

    if np is not None:
        np.random.seed(seed)

    if torch is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True

    if tf is not None:
        tf.random.set_seed(seed)

    if show:
        print("\nRandom seeds have been fixed as {}\n".format(seed))


def gen_ID(N=8):
    import random
    import string
    from datetime import datetime

    now = datetime.now()
    # now_str = now.strftime('%Y-%m-%d-%H:%M')
    today_str = now.strftime("%Y-%m%d")
    randlst = [random.choice(string.ascii_letters + string.digits) for i in range(N)]
    rand_str = "".join(randlst)
    return today_str + "_" + rand_str
    return rand_str


def gen_timestamp():
    from datetime import datetime

    now = datetime.now()
    now_str = now.strftime("%Y-%m%d-%H%M")
    return now_str


class Tee(object):
    """Example:
    import sys

    sys.stdout = Tee(sys.stdout, "stdout.txt")
    sys.stderr = Tee(sys.stderr, "stderr.txt")

    print("abc") # stdout
    print(1 / 0) # stderr
    # cat stdout.txt
    # cat stderr.txt
    """

    def __init__(self, sys_stdout_or_stderr, spath):
        self._files = [sys_stdout_or_stderr, open(spath, "w")]

    def __getattr__(self, attr, *args):
        return self._wrap(attr, *args)

    def _wrap(self, attr, *args):
        def g(*a, **kw):
            for f in self._files:
                res = getattr(f, attr, *args)(*a, **kw)
            return res

        return g


def tee(sys, sdir=None):
    """
    import sys

    sys.stdout, sys.stderr = tee(sys)

    print("abc")  # stdout
    print(1 / 0)  # stderr
    """

    import inspect
    import os

    ####################
    ## Determines sdir
    ####################
    if sdir is None:
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = "/tmp/fake.py"
        spath = __file__

        _sdir, sfname, _ = mngs.general.split_fpath(spath)
        sdir = _sdir + sfname + "/log/"

    os.makedirs(sdir, exist_ok=True)

    spath_stdout = sdir + "stdout.log"
    spath_stderr = sdir + "stderr.log"
    os.makedirs(mngs.general.split_fpath(spath_stdout)[0], exist_ok=True)

    sys_stdout = Tee(sys.stdout, spath_stdout)
    sys_stderr = Tee(sys.stdout, spath_stderr)

    print(
        "\nStandard Output/Error are going to be logged in the followings: \n  - {}\n  - {}\n".format(
            spath_stdout, spath_stderr
        )
    )
    return sys_stdout, sys_stderr



#!/usr/bin/env python3

import inspect
import os

import mngs

# import git


################################################################################
## PATH
################################################################################
def mk_spath(sfname, makedirs=False):

    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = "/tmp/fake.py"

    ## spath
    fpath = __file__
    fdir, fname, _ = split_fpath(fpath)
    sdir = fdir + fname + "/"
    spath = sdir + sfname

    if makedirs:
        os.makedirs(mngs.general.split_fpath(spath)[0], exist_ok=True)

    return spath


def find_the_git_root_dir():
    repo = git.Repo(".", search_parent_directories=True)
    return repo.working_tree_dir


def split_fpath(fpath):
    """Split a file path to (1) the directory path, (2) the file name, and (3) the file extention
    Example:
        dirname, fname, ext = split_fpath('../data/01/day1/split_octave/2kHz_mat/tt8-2.mat')
        print(dirname) # '../data/01/day1/split_octave/2kHz_mat/'
        print(fname) # 'tt8-2'
        print(ext) # '.mat'
    """
    dirname = os.path.dirname(fpath) + "/"
    base = os.path.basename(fpath)
    fname, ext = os.path.splitext(base)
    return dirname, fname, ext
